/*
    Dynamische CSS, vooral gericht op door voorkeur
    van bezoekers, via cookies onthouden, gestuurde
    kleuren.

    Dynamic CSS, specially targeted to colour schemes
    determined by user preference, remembered using
    cookies.

    Copyright 2002, 2013-2014, R.Harmsen

    3 February 2023:
    If no url is speficied, no CSS is fetched, but cookies
    are still set, and color correction (specifying dull
    colours) is done if necessary. This makes it possible
    to CSS from files directly, then call colschem.cgi
    only once, for the correction. Much for efficient.

*/

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/stat.h>
#include <time.h>

#include "../cgitools/cgitools.h"

static FILE *GetCssFile (void);
static int CheckCookie (void);
static int FilterColours(char *LineBuf);

char *ContentLength = "CONTENT_LENGTH";
extern char **environ;

static int VariableInTime (void);

int main (void)
{
   char LineBuf[1024];
   FILE *CssFp;
   int FilterWanted;

   FilterWanted = CheckCookie();

   printf("Content-type: text/css\n\n");

   if ((CssFp = GetCssFile()) != NULL)
   {
      while (fgets(LineBuf, sizeof LineBuf, CssFp) == LineBuf)
      {
         if (FilterWanted)
         {
            /* Filter does nothing, append to override instead */
            FilterColours(LineBuf);
         }
         fputs(LineBuf, stdout);
      }
      fclose(CssFp);
   }

   if (FilterWanted)
   {
      printf(
         "body,p,td,h1,h2,h3,h4,h5,h6,input,textarea,.citaat1,.citaat2,.citaat3,.citaat4 \n"
         "          {color: #000000; background-color: #ffffff; } \n"
         "a:link    {color: #0000c9; background-color: #ffffff; } \n"
         "a:visited {color: #004070; background-color: #ffffff; } \n"
         "a:active  {color: #0000ff; background-color: #aaaaaa; } \n"
         "a:hover   {color: #0000ff; background-color: #aaaaaa; } \n"
         );
      printf(".langmark {background-color:  #c0c0c0; } \n");
   }

   return 0;
}

/***********************************************************************
   Interpret command line argument of called CGI prog.
   Return a file pointer to the CSS file, if found, else
   return NULL
 ***********************************************************************/
static FILE *GetCssFile (void)
{
   FILE *fp = NULL;
   char *query;
   char **Vars = NULL, **V;

   query = getenv("QUERY_STRING");
   if (!query)
   {
      return NULL;
   }
   else
   {
      Vars = CgiConvVars(query, NULL);

      for (V = Vars; V[0]; V += 2)
      {
         if (strcmp(V[0], "url") == 0)
         {
            char URL[128];
            sprintf(URL, "css/%.*s", (int)((sizeof URL) - 4 - 1), V[1]);
            fp = fopen(URL, "r");
            break;
         }
      }
   }
   return fp;
}


/***********************************************************************

 ***********************************************************************/
static int CheckCookie (void)
{
   if (getenv("HTTP_COOKIE"))
   {
      char **Vars = NULL, **V;

      Vars = CgiConvCookies(getenv("HTTP_COOKIE"), NULL);
      for (V = Vars; V[0]; V += 2)
         if (strcmp(V[0], "NoWeirdColours") == 0)
            return atoi(V[1]);
   }

   // No cookie, no preference:
   // Do not filter, that is, show the weird, flashy colours
   // (changed back again 31 March 2012)
   // (25 June 2012: neutral colours as default again)
   // (12 July 2013: made it variable: flashy & weird some 15% of the time
   //  otherwise dull and neutral.
   return VariableInTime();
}


/***********************************************************************

 ***********************************************************************/
static int FilterColours(char *LineBuf)
{
   /* Filter does nothing, append to override instead */

   #if 0
   char *p;

   while ((p = strstr(LineBuf, "#000000")) != NULL)
   {
      memcpy(p, "#aaaa00", 7);
   }
   #endif

   return 0;
}

/***********************************************************************
   This function was added 12 July 2013.

   The idea is that those who did not click a link to explicitly ask for
   either weird or dull colours, will get weird colours some 33% of the
   time, and dull colours the remaining 67%. This is switched in an
   approx. 7 hour scheme, so it traverses slowly through night and day.
   Exact switchover times are severely randomised to get a surprise
   element into it.

   To make it efficient, the program does not read data from the status
   files WeirdOnn and WeirdOff, but instead only enquires (using the
   'stat' function) about their latest modification date. That's why
   there are two different status files, instead of only one with a
   yes/no content.
   The file that was most recently updated/recreated ("touched")
   determines which colour scheme the website visitor will get to see.

   Users who clicked a colour link and had a cookie set, will not be
   affected by all this. They always just get the colour scheme they
   prefer and asked for.
 ***********************************************************************/

static char *WeirdOnnFilename = "colschem/WeirdOnn";
static char *WeirdOffFilename = "colschem/WeirdOff";

#define DRAND() ((double)random() / (double)0x7fffffffL)
#define SRAND() (srandom((unsigned long)(getpid() + getppid() + time(NULL))))

/* Parameters, all times in seconds */

#define FractionWeird 0.85
#define RandomFactor  0.35
#define SwitchPeriod (13.5*(double)60*60)
#define CoolingOff (SwitchPeriod * FractionWeird * (1.0 - RandomFactor))

static int VariableInTime (void)
{
   struct stat WeirdOnnStat, WeirdOffStat;
   time_t TimeNow;
   FILE *fp;
   int Weird;

   if (stat(WeirdOnnFilename, &WeirdOnnStat) != 0 ||
       stat(WeirdOffFilename, &WeirdOffStat) != 0)
   {
      /* Initial create of two empty colour setting control files.
         They're empty because only the stat data (modification
         date) matters.
       */
      fp = fopen(WeirdOnnFilename, "w");
      if (fp) fclose(fp);
      fp = fopen(WeirdOffFilename, "w");
      if (fp) fclose(fp);
   }

   if (WeirdOnnStat.st_mtime >= WeirdOffStat.st_mtime)
   {
      Weird = 1; /* Don't filter, so display weird colours */
   }
   else
   {
      Weird = 0;
   }

   /* Is it time to switch over? */
   time(&TimeNow);

   if ((Weird && (TimeNow - WeirdOnnStat.st_mtime < CoolingOff)) ||
      (!Weird && (TimeNow - WeirdOffStat.st_mtime < CoolingOff)) )
   {
      ; /* Do nothing the first 20 or so minutes after a switch-over, to make
           it more efficient */
   }
   else
   {
      time_t TimeToSwitch;

      SRAND();

      if (Weird)
      {
         TimeToSwitch = FractionWeird * (float)SwitchPeriod;
         TimeToSwitch *= 1.0 + (RandomFactor * (DRAND() * 2 - 1.0));
         TimeToSwitch += WeirdOnnStat.st_mtime;
      }
      else
      {
         TimeToSwitch = (1.0 -  FractionWeird) * (float)SwitchPeriod;
         TimeToSwitch *= 1.0 + (RandomFactor * (DRAND() * 2 - 1.0));
         TimeToSwitch += WeirdOffStat.st_mtime;
      }

      // printf("/* TimeToSwitch %ld, TimeNow %ld */\n", TimeToSwitch, TimeNow);

      if (TimeToSwitch <= TimeNow)
      {
         fp = fopen(Weird ? WeirdOffFilename : WeirdOnnFilename, "w");
         if (fp) fclose(fp);
         Weird = !Weird;
      }
   }

   return !Weird; /* Weird = don't filter, Dull = do filter, hence the ! */
}
